"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import logging

from tornado.options import define, options
import tornado.httpserver
import tornado.ioloop
import tornado.web

from .views import IndexHandler, SetupHandler, RunHandler

# Define tornado options
define(
    "worker_home",
    default="/tmp/bad-framework",
    help="Working directory for the worker process.",
)
define("worker_port", default=3291, help="Port for the worker process.")
define("worker_debug", default=False, help="Activate development mode.")

# Define our own logging configuration
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)5s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
log = logging.getLogger("bad.server.worker")


class BADWorkerServer:
    def __init__(self, port, home_dir, debug):
        self._port = int(port)
        self._app = tornado.web.Application(
            [
                (r"/", IndexHandler),
                (r"/index.html", IndexHandler),
                (r"/setup/", SetupHandler),
                (r"/run/", RunHandler),
            ],
            debug=debug,
            worker_port=port,
            worker_home=home_dir,
        )
        self._server = tornado.httpserver.HTTPServer(
            self._app, max_buffer_size=524288000
        )

    def start(self):
        log.info(">>> Starting BAD worker on port %d", self._port)
        self._server.bind(self._port)
        self._server.start(1)
        tornado.ioloop.IOLoop.current().start()


def main():
    tornado.options.parse_command_line()

    bad_worker = BADWorkerServer(
        options.worker_port, options.worker_home, options.worker_debug
    )
    bad_worker.start()


if __name__ == "__main__":
    main()



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import itertools
import numpy as np
import sys
import subprocess

from . import adt


def generate_experiments_settings(datasets, parameter_settings):
    """Generates experiment settings for a particular experiment suite.
    The total number of experiment settings depends on the number of required data sets
    and on the parameter configurations.

    :param datasets: (list[str]) dataset names to be used for the experiments.
    :param parameter_settings: (dict) a map between parameter names
    and parameter values.
    :return: iterable of ExperimentSettings.
    """
    parameter_names = parameter_settings.keys()
    parameter_strings = []
    for parameter_values in get_parameter_combinations(parameter_settings):
        parameter_dict = dict(
            (name, value) for name, value in zip(parameter_names, parameter_values)
        )
        parameter_strings.append(get_parameter_string(parameter_dict))
    for combination in itertools.product(datasets, parameter_strings):
        yield adt.ExperimentSettings(
            dataset_name=combination[0],
            parameters=combination[1],
        )


def get_parameter_combinations(parameters):
    """Takes a dictionary of parameters and returns all possible parameter combinations.

    :param parameters: (dict) dictionary name -> ValueParameter or RangeParameter.
    :return: (iterator) N-tuple iterator corresponding to parameter combinations.
    N is the number of items in parameters.
    """
    distinct_values = []
    for param_name, param in parameters.items():
        param_type = param.__class__
        if param_type == adt.ValueParameter:
            distinct_values.append([param.value])
        elif param_type == adt.RangeParameter:
            value_range = [
                boxed_value.item()
                for boxed_value in np.arange(
                    start=param.start, stop=param.end + param.step, step=param.step
                )
            ]
            distinct_values.append(value_range)
        else:
            raise ValueError("invalid parameter type '{}'".format(param_type))
    # Generate all parameters combinations
    return itertools.product(*distinct_values)


def get_parameter_string(parameters):
    """Converts a parameter dictionary into a parameter string.

    :param parameters: (dict) parameters dictionary.
    :return: (string) formatted parameter string.

    Examples:

    >>> get_parameter_string({"a": 10, "b": "hello"})
    'a=10;b=hello'
    >>> get_parameter_string({})
    ''
    """
    return ";".join(
        "{key}={value}".format(key=k, value=v) for k, v in parameters.items()
    )


def init_worker_environments(workers, suite_id, candidate_id, datasets):
    for worker in workers:
        message = {
            "master_address": worker.master_address,
            "suite_id": suite_id,
            "candidate_id": candidate_id,
            "datasets": [dataset.name for dataset in datasets],
        }
        worker.session.post_json("setup/", message)


def install_requirements(requirements):
    """Installs a requirements.txt file using the pip defined in the current interpreter.

    :param requirements: (list[string]) list of pip requirement specifiers.
    :return: None
    """
    subprocess.check_call([sys.executable, "-m", "pip", "install", *requirements])


def load_data_matrix(path):
    """Read a data file in ARFF format and returns
    the numpy matrix corresponding to the data.

    :param path: (string) path to data file.
    :return: (numpy.ndarray) data matrix
    """
    return np.loadtxt(path, dtype=float, comments=["#", "@"], delimiter=",")


def load_parameter_string(parameter_string):
    """Converts a parameter string into a dictionary parameter.
    Parameter values are represented as strings.

    :param parameter_string: (string) parameter string.
    :return: (dict) parameters dictionary.

    Examples:

    >>> load_parameter_string("a=10;b=hello")
    {'a': '10', 'b': 'hello'}
    >>> load_parameter_string("")
    {}
    """
    parameters_dict = {}
    if parameter_string:
        parameters = parameter_string.split(";")
        for key_value in parameters:
            k, v = key_value.split("=")
            parameters_dict[k] = v
    return parameters_dict



import pytest

from bad_framework.bad_client.monitor import (
    get_count_by_status,
    initialize_status_cache,
    update_status_cache,
)


def test_get_count_by_status():
    test_cache = {
        "dummy_exp_1": "created",
        "dummy_exp_2": "scheduled",
        "dummy_exp_3": "scheduled",
    }
    assert 1 == get_count_by_status(test_cache, "created")
    assert 2 == get_count_by_status(test_cache, "scheduled")


def test_get_count_by_status_when_cache_is_none():
    with pytest.raises(ValueError):
        get_count_by_status(None, "created")


def test_get_count_by_status_with_invalid_status():
    with pytest.raises(ValueError):
        test_cache = {
            "dummy_exp_1": "created",
            "dummy_exp_2": "scheduled",
        }
        get_count_by_status(test_cache, "invalid_status")


def test_initialize_status_cache_with_none():
    with pytest.raises(ValueError):
        initialize_status_cache(None)


def test_initialize_status_cache():
    example_response = [
        {"id": "dummy_exp_1", "status": "running"},
        {"id": "dummy_exp_2", "status": "scheduled"},
    ]
    expected_cache = {
        "dummy_exp_1": "running",
        "dummy_exp_2": "scheduled",
    }
    assert expected_cache == initialize_status_cache(example_response)


def test_update_status_cache():
    test_cache = {
        "dummy_exp_1": "created",
        "dummy_exp_2": "scheduled",
    }
    test_experiments = [
        {"id": "dummy_exp_1", "status": "created"},
        {"id": "dummy_exp_2", "status": "running"},
    ]
    expected_cache = {
        "dummy_exp_1": "created",
        "dummy_exp_2": "running",
    }
    update_status_cache(test_experiments, test_cache)
    assert expected_cache == test_cache


def test_get_suite_experiments():
    # example_response = {
    #     "suite_id": "dummy_suite",
    #     "experiments": [
    #         {"id": "dummy_exp_1", "status": "dummy_status_1"},
    #         {"id": "dummy_exp_2", "status": "dummy_status_2"},
    #     ],
    # }
    pass



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""


class InvalidCandidateError(Exception):
    pass


class SchedulingError(Exception):
    pass


def get_response_message(status, payload):
    return {
        "status": int(status),
        **payload,
    }


def get_error_message(reason):
    return get_response_message(500, {"error": reason})



import pytest


@pytest.fixture(scope="function")
def bad_caplog(caplog):
    """Uses the pytest log capturing fixture to capture the
    application's log messages up to the DEBUG level.

    :param caplog (pytest.logging.LogCaptureFixture): log capturing fixture
    :return: (pytest.logging.LogCaptureFixture): modified log capturing fixture
    """
    import logging

    caplog.set_level(level=logging.DEBUG, logger="bad.client")
    return caplog



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
from sklearn.neighbors import LocalOutlierFactor


class LOF:
    """Implements the Local Outlier Factor (LOF) outlier detection method proposed in

    Breunig, Markus M., et al. "LOF: identifying density-based local outliers."
    ACM sigmod record 29.2 (2000): 93-104.

    Parameters:
    - k: (int) number of neighbors to consider (defaults to 10). Must be smaller than
    the data set size.
    - seed: (int) random number generator seed.

    This class wraps the implementation provided by the scikit-learn library
    (https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.LocalOutlierFactor.html).  # noqa 501
    """

    def __init__(self, **kwargs):
        self.k = int(kwargs.get("k", 10))
        self._model = None

    def fit(self, train_data):
        self._model = LocalOutlierFactor(
            n_neighbors=self.k,
            algorithm="kd_tree",
            contamination="auto",
            novelty=True,
        ).fit(train_data)
        return self

    def score(self, element):
        if not self._model:
            raise ValueError("invalid state. The model has not been trained.")
        element = element.reshape(1, -1)
        return self._model.decision_function(element)[0]



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import json
import importlib.util
import logging
import os
import traceback

from sklearn.metrics import (
    average_precision_score,
    roc_auc_score,
    roc_curve,
)
import jinja2 as j2
import matplotlib.pyplot as plt
import numpy as np
import tornado.web

from bad_framework.bad_utils import (
    install_requirements,
    load_parameter_string,
    load_data_matrix,
)
from bad_framework.bad_utils.adt import ExperimentStatus
from bad_framework.bad_utils.files import get_candidate_name
from bad_framework.bad_utils.magic import LOG_FORMAT
from bad_framework.bad_utils.network import AsyncHTTPSessionManager

logging.basicConfig(format=LOG_FORMAT, level="INFO")
log = logging.getLogger("bad.server.worker")


class BaseWorkerHandler(tornado.web.RequestHandler):
    def write_error(self, status_code, **kwargs):
        self.write("<div>You've just got a big, BAD server error... sorry :(</div>")

    def get_wd(self):
        return self.application.settings["worker_home"]

    def head(self):
        self.set_status(200)
        self.flush()
        self.finish()


class IndexHandler(BaseWorkerHandler):
    def get(self):
        try:
            jinja2_env = j2.Environment(
                loader=j2.PackageLoader("bad_framework.bad_worker", "templates")
            )
            template = jinja2_env.get_template("index.html")
            rendered_template = template.render()
            self.set_status(200)
            self.set_header("Content-type", "text/html")
            self.write(rendered_template.encode())
        finally:
            self.flush()
            self.finish()


class SetupHandler(BaseWorkerHandler):
    async def _setup_worker(self):
        log.info(">>> Setting up worker")
        message = json.loads(self.request.body)

        candidate_id = message["candidate_id"]
        datasets = message["datasets"]
        master_address = message["master_address"]
        requirements = message["requirements"]
        suite_id = message["suite_id"]

        base_path = self.get_wd()
        master_session = AsyncHTTPSessionManager(master_address)
        await self._download_candidate_files(
            master_session, candidate_id, suite_id, base_path
        )
        install_requirements(requirements)
        await self._download_datasets(master_session, base_path, datasets)
        log.info("<<< Done.")

    @classmethod
    async def _download_candidate_files(
        cls, master_session, candidate_id, suite_id, base_path
    ):
        """Downloads candidate source file and its requirements.txt file.
        Installs the requirements.txt file in the current interpreter.

        :param master_session: (bad_utils.network.SessionManager) master session manager
        :param suite_id: (str) suite id
        :param base_path: (str) path to worker home directory
        :return: None
        """
        log.info("Downloading candidate file...")
        await master_session.download_file(
            url="candidate/{candidate_id}/".format(candidate_id=candidate_id),
            path="{home_dir}/{suite_id}/candidate.py".format(
                home_dir=base_path, suite_id=suite_id
            ),
        )

    @classmethod
    async def _download_datasets(cls, master_session, base_path, datasets):
        """Downloads the data sets required for suite execution.

        Only downloads data sets that have not been previously downloaded.

        :param master_session: (bad_utils.network.SessionManager) master session manager
        :param datasets: (list[string]) list of data set names to download.
        :return: None
        """
        log.info("Downloading datasets: %r", datasets)
        data_urls = [
            "dataset/{dataset_name}/".format(dataset_name=dataset)
            for dataset in datasets
        ]
        data_paths = [
            "{base_path}/datasets/{dataset_name}.arff".format(
                base_path=base_path, dataset_name=dataset
            )
            for dataset in datasets
        ]
        for data_url, data_path in zip(data_urls, data_paths):
            if not os.path.isfile(data_path):
                await master_session.download_file(data_url, data_path)

    async def post(self):
        try:
            await self._setup_worker()
            self.set_status(status_code=200)
        except Exception:
            log.error(traceback.format_exc())  # Write on worker log
        finally:
            await self.finish()


class RunHandler(BaseWorkerHandler):
    """Handles POST requests to the "/run/" path.

    This executes BAD experiment.
    """

    def prepare(self):
        message = json.loads(self.request.body)
        self._data_name = message["data_name"]
        self._experiment_id = message["experiment_id"]
        self._home_dir = self.get_wd()
        self._master_address = message["master_address"]
        self._master_session = AsyncHTTPSessionManager(self._master_address)
        self._suite_id = message["suite_id"]
        self._parameters = load_parameter_string(message["parameters"])
        self._trainset_size = self._parameters["trainset_size"]

    async def _update_experiment_status(self, status):
        experiment_url = "experiment/{experiment_id}/".format(
            experiment_id=self._experiment_id
        )
        message = {"status": status}
        await self._master_session.post_json(url=experiment_url, data=message)

    @classmethod
    def _import_candidate_module(cls, candidate_path):
        spec = importlib.util.spec_from_file_location("bad_candidate", candidate_path)
        bad_candidate = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(bad_candidate)
        return bad_candidate

    def _generate_metrics_file(self, labels, scores):
        metrics_path = "{home_dir}/{suite_id}/{experiment_id}/metrics.json".format(
            home_dir=self._home_dir,
            suite_id=self._suite_id,
            experiment_id=self._experiment_id,
        )
        if not os.path.exists(os.path.dirname(metrics_path)):
            os.makedirs(os.path.dirname(metrics_path))

        roc_auc = roc_auc_score(y_score=scores, y_true=labels)
        average_precision = average_precision_score(y_score=scores, y_true=labels)

        metrics = {
            "experiment_id": self._experiment_id,
            "roc_auc": roc_auc,
            "average_precision": average_precision,
        }
        pretty_json = json.dumps(metrics, indent=2)
        with open(metrics_path, "w") as metrics_file:
            metrics_file.write(pretty_json)
        return metrics_path

    def _generate_scores(
        self,
        candidate,
        data_matrix,
    ):
        feature_matrix = data_matrix[:, 2:]

        # Subsample feature matrix to obtain training set
        np.random.seed(int(self._parameters["seed"]))
        num_rows = feature_matrix.shape[0]
        trainset_size = int(num_rows * float(self._parameters["trainset_size"]))
        training_indexes = np.random.choice(num_rows, size=trainset_size, replace=False)
        training_matrix = feature_matrix[training_indexes, :]

        candidate = candidate.fit(training_matrix)
        scores = np.apply_along_axis(candidate.score, axis=1, arr=feature_matrix)
        return scores

    def _generate_roc_file(self, labels, scores):
        roc_path = "{home_dir}/{suite_id}/{experiment_id}/roc.png".format(
            home_dir=self._home_dir,
            suite_id=self._suite_id,
            experiment_id=self._experiment_id,
        )
        if not os.path.exists(os.path.dirname(roc_path)):
            os.makedirs(os.path.dirname(roc_path))
        fpr, tpr, _ = roc_curve(y_score=scores, y_true=labels)
        plt.figure()
        plt.xlabel("False Positive Rate")
        plt.ylabel("True Positive Rate")
        plt.plot([0, 1], [0, 1], color="navy", linestyle="--")
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.title("ROC")
        plt.plot(fpr, tpr, color="blue", lw=2)
        plt.savefig(roc_path, format="png")
        plt.close()
        return roc_path

    async def _send_results(
        self,
        # scores_path,
        metrics_path,
        roc_path,
    ):
        results_url = "experiment/{experiment_id}/results/".format(
            experiment_id=self._experiment_id,
        )

        with open(metrics_path, "rb") as metrics_file:
            metrics_content = metrics_file.read()

        with open(roc_path, "rb") as roc_file:
            roc_content = roc_file.read()

        files = {
            # "scores": open(scores_path, "rb"),
            "metrics.json": metrics_content,
            "roc.png": roc_content,
        }
        await self._master_session.post_files(url=results_url, files=files)

    async def _run_experiment(self):
        log.info(
            ">>> Loading experiment %s received from %s",
            self._experiment_id,
            self._master_address,
        )
        home_dir = self.get_wd()
        candidate_path = "{home_dir}/{suite_id}/candidate.py".format(
            home_dir=home_dir, suite_id=self._suite_id
        )
        data_path = "{home_dir}/datasets/{data_name}.arff".format(
            home_dir=home_dir,
            data_name=self._data_name,
        )
        candidate_name = get_candidate_name(candidate_path)

        log.info("Loading candidate %s from candidate module...", candidate_name)
        bad_candidate = self._import_candidate_module(candidate_path)
        candidate_class = getattr(bad_candidate, candidate_name)
        candidate = candidate_class(**self._parameters)

        try:
            log.info(">>> Running experiment %s", self._experiment_id)
            log.info("Loading data matrix from %s", data_path)
            data_matrix = load_data_matrix(data_path)
            labels = data_matrix[:, 1]
            log.info("Running experiment...")
            scores = self._generate_scores(candidate, data_matrix)
            log.info("Experiment completed.")
            # log.info("Saving scores file...")
            # scores_path = _save_scores_to_file(
            #     suite_id=suite_id,
            #     experiment_id=experiment_id,
            #     result_matrix=result_matrix,
            # )
            log.info("Generating metrics file...")
            metrics_path = self._generate_metrics_file(
                labels=labels,
                scores=scores,
            )
            log.info("Generating ROC plot...")
            roc_path = self._generate_roc_file(
                labels=labels,
                scores=scores,
            )
            log.info("Sending results to master...")
            await self._send_results(
                # scores_path=scores_path,
                metrics_path=metrics_path,
                roc_path=roc_path,
            )
            log.info("<<< Experiment %s completed successfully.", self._experiment_id)
        except Exception as e:
            await self._update_experiment_status(status=ExperimentStatus.FAILED)
            log.error("Experiment runtime error: %s", e)
            log.error("<<< Experiment %s failed.", self._experiment_id)
            raise e  # Handled in self post()

    async def post(self):
        try:
            await self._update_experiment_status(status=ExperimentStatus.RUNNING)
            await self._run_experiment()
            self.set_status(status_code=200)
        except Exception:
            log.error(traceback.format_exc())  # Write on worker log
        finally:
            await self.finish()



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
from pyod.models.loci import LOCI as Loci


class LOCI:
    """Implements the Local Outlier Correlation Integral (LOCI) method proposed in:

    Papadimitriou, Spiros, et al. "Loci: Fast outlier detection using
    the local correlation integral." Proceedings 19th international
    conference on data engineering (Cat. No. 03CH37405). IEEE, 2003.

    Parameters:
    - alpha: (float) local neighborhood selection parameter (defaults to 0.5).
    - k: (int) outlier cutoff threshold (defaults to 3).
    - seed: (int) random number generator seed.

    This class wraps the implementation provided by the PyOD library
    (see https://pyod.readthedocs.io/en/latest/pyod.models.html#module-pyod.models.loci).  # noqa 501
    """

    def __init__(self, **kwargs):
        self.alpha = float(kwargs.get("alpha", 0.5))
        self.k = int(kwargs.get("k", 3))
        self.seed = int(kwargs["seed"])
        self._model = None

    def fit(self, train_data):
        self._model = Loci(
            alpha=self.alpha,
            k=self.k,
        ).fit(train_data)
        return self

    def score(self, element):
        if not self._model:
            raise ValueError("invalid state. The model has not been trained.")
        element = element.reshape(1, -1)
        return self._model.decision_function(element)[0]



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.

Main module for the BAD master process implementation.
"""
from time import sleep
import json
import logging
import os
import traceback

from jinja2 import Environment, PackageLoader
from tornado.ioloop import IOLoop
import tornado.web

from bad_framework.bad_candidates import (
    get_candidate_requirements,
    get_default_candidate,
)
from bad_framework.bad_utils import generate_experiments_settings, load_parameter_string
from bad_framework.bad_utils.adt import (
    CandidateSpec,
    DataSpec,
    ExperimentStatus,
    RangeParameter,
    ValueParameter,
)
from bad_framework.bad_utils.errors import get_response_message, get_error_message
from bad_framework.bad_utils.files import (
    get_candidate_filename,
    get_include_dir,
    save_file,
)
from .models import Candidate, Dataset, Experiment, Suite, Worker

log = logging.getLogger("bad.server.master")


class BaseMasterHandler(tornado.web.RequestHandler):
    """Base handler for BAD master HTTP requests."""

    def write_error(self, status_code, **kwargs):
        """Writes a custom error page to the output stream.

        :param status_code: (int) HTTP status code (not used).
        :param kwargs: (dict) keyword arguments (not used).
        """
        self.write("<div>You just got a big, BAD server error... sorry :(</div>")

    def get_wd(self):
        """Returns the path to the working directory for the BAD master.

        :return: (string) path to working directory.
        """
        return self.application.settings["master_home"]

    def head(self):
        """Handler for HTTP HEAD method."""
        self.set_status(200)
        self.flush()
        self.finish()

    def get_file_contents(self, file_name):
        """Returns the body of a multi-part encoded file sent with the HTTP request.

        The file is identified by a filename set by the HTTP client.

        :param file_name: (string) identifier of the file to fetch.
        :return: (bytes) contents of the file.
        """
        return self.request.files[file_name][0].body


class IndexHandler(BaseMasterHandler):
    """Handler for the BAD master index page."""

    def _render_index(self):
        """Renders the index HTML page."""
        suites = Suite.get_all()
        jinja2_env = Environment(
            loader=PackageLoader("bad_framework.bad_master", "templates")
        )
        template = jinja2_env.get_template("index.html")
        rendered_template = template.render(suites=suites)
        self.set_header("Content-type", "text/html")
        self.write(rendered_template.encode())

    def get(self):
        """Handler for HTTP GET method.

        Renders the index page and returns.
        """
        try:
            self._render_index()
            self.set_status(200)
        except Exception as e:
            self.set_status(500, reason=str(e))
        finally:
            self.flush()


class CandidateHandler(BaseMasterHandler):
    """Handler for candidate related requests."""

    def _get_candidate_file(self, candidate_id):
        """Writes the candidate.py file contents to the output stream.

        :param candidate_id: (string) candidate id.
        """
        candidate = Candidate.get_by_id(candidate_id)
        with open(candidate.source, "rb") as candidate_file:
            content = candidate_file.read()
        return content

    def get(self, candidate_id):
        """Handler for HTTP GET method.

        Returns candidate files depending on the requested path.
        """
        try:
            # if "requirements" in self.request.path:
            #     self._get_requirements_file(candidate_id)
            # else:
            response_content = self._get_candidate_file(candidate_id)
            self.write(response_content)
            self.set_status(200)
        except Exception as e:
            log.error(traceback.format_exc())
            message = get_error_message(reason=str(e))
            self.write(message)
        finally:
            self.flush()
            self.finish()


class DatasetHandler(BaseMasterHandler):
    """Handler for dataset related requests."""

    def get(self, dataset_name):
        """Handler for HTTP GET method.

        Writes the requested dataset to the output stream.

        :param dataset_name: (string) dataset name.
        """
        try:
            dataset = Dataset.get_by_name(dataset_name)
            if dataset:
                # Default dataset
                dataset_path = dataset.path
            else:
                # Locally uploaded dataset
                dataset_path = os.path.join(
                    self.get_wd(), "dataset", dataset_name + ".arff"
                )
            with open(dataset_path, "rb") as data_file:
                content = data_file.read()
                self.write(content)
            self.set_status(200)
        except Exception as e:
            log.error(traceback.format_exc())  # Write on master log
            self.set_status(status_code=500, reason=str(e))
        finally:
            self.flush()
            self.finish()


class ExperimentHandler(BaseMasterHandler):
    """Handler for experiment related requests."""

    def _render_experiment_details(self, experiment_id):
        """Renders the experiment details HTML page.

        :param experiment_id: (string) experiment id.
        """
        experiment = Experiment.get_by_id(experiment_id)
        candidate = Candidate.get_by_id(experiment.candidate)
        jinja2_env = Environment(
            loader=PackageLoader("bad_framework.bad_master", "templates")
        )
        template = jinja2_env.get_template("experiment_details.html")
        rendered_template = template.render(candidate=candidate, experiment=experiment)
        self.set_header("Content-type", "text/html")
        self.write(rendered_template.encode())

    def _update_experiment_status(self, experiment_id):
        """Updates the status of an experiment.

        The updated status is encoded in the JSON payload of the request.

        :param experiment_id: (string) experiment id.
        """
        message = json.loads(self.request.body)
        experiment = Experiment.get_by_id(experiment_id)
        experiment.update_status(message["status"])

    def get(self, experiment_id):
        """Handler for HTTP GET method.

        Renders the experiment details page.

        :param experiment_id: (string) experiment id.
        """
        try:
            self._render_experiment_details(experiment_id)
            self.set_status(200)
        except Exception as e:
            log.error(traceback.format_exc())
            self.set_status(status_code=500, reason=str(e))
        finally:
            self.flush()
            self.finish()

    def post(self, experiment_id):
        """Handler for HTTP POST method.

        Updates an experiment's status.

        :param experiment_id: (string) experiment id.
        """
        try:
            self._update_experiment_status(experiment_id)
            self.set_status(200)
        except Exception as e:
            log.error(traceback.format_exc())
            self.set_status(500, reason=str(e))
        finally:
            self.finish()


class ResultHandler(BaseMasterHandler):
    """Handler for results related requests."""

    def save_experiment_results(self, experiment_id):
        """Saves the results experiment results received from a worker in a local file.
        Updates the experiment with the paths to the result files.

        :param experiment_id: (string) experiment id.
        """
        log.info("Saving results for experiment %s", experiment_id)
        experiment = Experiment.get_by_id(experiment_id)

        base_path = "{home_dir}/{suite_id}/{experiment_id}/".format(
            home_dir=self.get_wd(),
            suite_id=experiment.suite,
            experiment_id=experiment_id,
        )

        # If worker and master are on the same host, the files are already there
        metrics_path = base_path + "metrics.json"
        if not os.path.exists(metrics_path):
            metrics_content = self.get_file_contents("metrics.json")
            save_file(metrics_content, metrics_path)

        roc_path = base_path + "roc.png"
        if not os.path.exists(roc_path):
            roc_content = self.get_file_contents("roc.png")
            save_file(roc_content, roc_path)

        experiment.metrics = metrics_path
        experiment.roc = roc_path
        experiment.update_status(ExperimentStatus.COMPLETED)

    def post(self, experiment_id):
        """Handler for HTTP POST method.

        Receives the experiment result files from a worker and saves them to disk.

        :param experiment_id: (string) experiment id.
        """
        try:
            self.save_experiment_results(experiment_id)
            self.set_status(status_code=200)
        except Exception as e:
            log.error(traceback.format_exc())
            self.set_status(500, reason=str(e))
        finally:
            self.finish()

    def _get_metrics_file(self, experiment_id):
        """Writes the metric.json file contents to the output stream.

        :param experiment_id: (string) experiment id.
        """
        experiment = Experiment.get_by_id(experiment_id)
        with open(experiment.metrics, "rb") as metrics_file:
            self.set_header("Content-type", "application/json")
            self.write(metrics_file.read())

    def _get_roc_file(self, experiment_id):
        """Writes the roc.png file contents to the output stream.

        :param experiment_id: (string) experiment id.
        """
        experiment = Experiment.get_by_id(experiment_id)
        with open(experiment.roc, "rb") as roc_file:
            self.set_header("Content-type", "image/png")
            self.write(roc_file.read())

    def get(self, experiment_id):
        """Handler for HTTP GET method.

        Returns the experiment's results file.

        :param experiment_id: (string) experiment id.
        """
        try:
            if "metrics.json" in self.request.path:
                self._get_metrics_file(experiment_id)
            elif "roc.png" in self.request.path:
                self._get_roc_file(experiment_id)
            self.set_status(200)
        except Exception as e:
            log.error(traceback.format_exc())
            self.set_status(500, reason=str(e))
        finally:
            self.flush()
            self.finish()


class SuiteHandler(BaseMasterHandler):
    """Handler for Suite related requests."""

    def _save_candidate_files(self, file_paths):
        """Writes candidate files to disc. The candidate files are
        multi-part encoded in the request payload.

        TODO:
         - refactor to remove files

        The files are written in the master working directory under the
        relative suite directory.

        :param file_paths: (dict) paths to candidate files.
        """
        save_file(self.get_file_contents("candidate_source"), file_paths["candidate"])
        save_file(
            self.get_file_contents("candidate_requirements"),
            file_paths["requirements"],
        )
        save_file(
            self.get_file_contents("candidate_parameters"), file_paths["parameters"]
        )

    @classmethod
    def _generate_suite_experiments(cls, suite, candidate, parameters, datasets):
        """Creates the experiments related to this suite.

        An experiment is created for each dataset/parameter value combination.

        :param suite: (models.Suite) suite object.
        :param candidate: (models.Candidate) candidate object.
        :param parameters: (dict) parameter specification dictionary.
        :param datasets: (list[models.Dataset]) list of dataset objects.
        :return: (list[models.Experiment]) list of created experiments
        """
        for experiment_setting in generate_experiments_settings(datasets, parameters):
            Experiment.create(
                suite_id=suite.id,
                candidate_id=candidate.id,
                dataset_name=experiment_setting.dataset_name,
                parameters=experiment_setting.parameters,
            )
        return Experiment.get_by_suite(suite.id)

    @classmethod
    async def _schedule_experiment(cls, experiment, worker):
        """Sends an experiment to be scheduled to a worker.

        :param worker: (models.Worker) worker object.
        :param experiment: (models.Experiment) experiment object.
        """
        message = {
            "suite_id": experiment.suite,
            "data_name": experiment.dataset,
            "experiment_id": experiment.id,
            "master_address": worker.master_address,
            "parameters": experiment.parameters,
        }
        experiment.update_status(ExperimentStatus.SCHEDULED)
        await worker.session.post_json("run/", message)

    @classmethod
    def _count_running_experiments(cls, experiments):
        running_experiments = [
            experiment
            for experiment in experiments
            if experiment.status == ExperimentStatus.RUNNING
        ]
        return len(running_experiments)

    @classmethod
    def _get_update_message(cls, running, scheduled, total):
        return "...{r} ({s}/{t}) running (scheduled/total) experiments.".format(
            r=running,
            s=scheduled,
            t=total,
        )

    async def _start_scheduling_loop(self, experiments, workers):
        """Starts the experiment scheduling loop. Experiments are scheduled
        in time in order not to overload the workers.

        Each worker runs one experiment at-a-time.

        TODO:
         - make more clear/elegant/split

        :param experiments: (list[models.Experiment]) list of experiments to schedule.
        :param workers: (list[models.Worker]) list of workers to
        schedule the experiment on.
        """
        workers_num = len(workers)
        experiments_num = len(experiments)

        log.info(">>> Starting scheduling loop")
        log.info("Found %d available workers.", workers_num)
        log.info("Running %d experiments.", experiments_num)

        scheduled_experiments = []
        todo_experiments = experiments
        worker_index = 0
        scheduled_experiments_num = 0

        while todo_experiments:
            running_experiments_num = self._count_running_experiments(
                experiments=scheduled_experiments
            )
            if running_experiments_num < workers_num:
                next_experiment, *todo_experiments = todo_experiments
                await self._schedule_experiment(
                    experiment=next_experiment,
                    worker=workers[worker_index],
                )
                scheduled_experiments.append(next_experiment)
                worker_index = (worker_index + 1) % workers_num
            sleep(0.1)
            scheduled_experiments_num = len(scheduled_experiments)
        log.info(
            "<<< Scheduling loop completed (%d/%d).",
            scheduled_experiments_num,
            experiments_num,
        )

    @classmethod
    async def _initialize_worker_envs(cls, suite_id, candidate, workers, datasets):
        """Initializes each worker's environment with the dependencies
        required to run the experiments.

        :param suite_id: (string) suite id.
        :param workers: (list[models.Worker]) list of workers to initialize.
        :param datasets: (list[string]) list of dataset names.
        for the experiments.
        """
        for worker in workers:
            log.info("Setting up worker at %s on port %d", worker.hostname, worker.port)
            message = {
                "master_address": worker.master_address,
                "suite_id": suite_id,
                "candidate_id": candidate.id,
                "requirements": candidate.requirements,
                "datasets": datasets,
            }
            response = await worker.session.post_json("setup/", message)
            if response.status_code == 200:
                log.info("worker initialized correctly.")
            else:
                raise ValueError("worker initialization failed: ", response.reason)

    def _load_candidate(self, candidate_name):
        candidate_dir = os.path.join(get_include_dir(), "candidates")
        candidate_file = os.path.join(candidate_dir, candidate_name + ".py")

        if not os.path.exists(candidate_file):
            raise ValueError("candidate '{}' does not exists.".format(candidate_name))

        # Copy candidate files to suite directory, return candidate object
        raise NotImplementedError()

    def _get_candidate(self, suite_id, message):
        candidate_spec = CandidateSpec(*message["candidate"])

        parameters = self._load_parameters_list(message["candidate_parameters"])

        if candidate_spec.source == "local":
            candidate_src = self.get_file_contents("candidate_source")
            requirements = message["candidate_requirements"]
            candidate_filename = get_candidate_filename(self.get_wd(), suite_id)
            save_file(candidate_src, candidate_filename)
        else:
            candidate_name = candidate_spec.url
            candidate_filename = get_default_candidate(candidate_name)
            requirements = get_candidate_requirements(candidate_name)
        return Candidate.create(
            suite_id=suite_id,
            source_filename=candidate_filename,
            parameters=parameters,
            requirements=requirements,
        )

    def _add_local_dataset(self, message):
        dataset_spec = DataSpec(*message["data"])

        if dataset_spec.source == "local":
            dataset_src = self.get_file_contents("data_source")
            dataset_basename = os.path.basename(dataset_spec.url).replace("_", "-")
            dataset_filename = os.path.join(self.get_wd(), "dataset", dataset_basename)
            save_file(dataset_src, dataset_filename)
            return dataset_basename.replace(".arff", "")
        return dataset_spec.url

    @classmethod
    def _load_parameters_list(cls, parameters_list):
        parameters = {}
        for param_tuple in parameters_list:
            if len(param_tuple) == 2:
                param_name, param_value = param_tuple
                parameters[param_name] = ValueParameter(param_value)
            elif len(param_tuple) == 4:
                param_name, start, end, step = param_tuple
                parameters[param_name] = RangeParameter(start, end, step)
            else:
                raise ValueError(
                    "invalid parameter specification: '{}'".format(param_tuple)
                )
        return parameters

    async def post(self):
        """Handler for HTTP POST method.

        Creates a new experiment suite with the settings encoded in the request's payload.
        """
        try:
            message = json.loads(self.get_file_contents("suite_settings"))

            suite = Suite.create()

            data_name = self._add_local_dataset(message)
            if not Dataset.get_all():
                Dataset.setup()

            if data_name:
                dataset_names = [data_name]
            else:
                dataset_names = [dataset.name for dataset in Dataset.get_all()]

            master_address = message["master_address"]
            workers_list = message["workers"]
            if not Worker.get_all():
                Worker.setup(workers_list, master_address)
            workers = list(
                Worker.get_all()  # this returns a non-subscriptable set-like
            )

            candidate = self._get_candidate(suite.id, message)

            experiments = self._generate_suite_experiments(
                suite=suite,
                candidate=candidate,
                datasets=dataset_names,
                parameters=candidate.parameters,
            )
            await self._initialize_worker_envs(
                suite.id, candidate, workers, dataset_names
            )

            # runs scheduling loop in the background
            IOLoop.current().add_callback(
                callback=self._start_scheduling_loop,
                experiments=experiments,
                workers=workers,
            )

            message = get_response_message(status=200, payload={"suite_id": suite.id})
            self.set_status(200)
            self.write(message)
        except Exception as e:
            log.error(traceback.format_exc())
            message = get_error_message(reason=str(e))
            self.write(message)
        finally:
            await self.finish()

    def _render_suite_details(self, suite_id):
        """Renders the suite details HTML page.

        :param suite_id: (string) suite id.
        """
        candidate = Candidate.get_by_suite(suite_id)
        experiments = Experiment.get_by_suite(suite_id)
        jinja2_env = Environment(
            loader=PackageLoader("bad_framework.bad_master", "templates")
        )
        template = jinja2_env.get_template("suite_details.html")
        rendered_template = template.render(
            candidate=candidate, experiments=experiments
        )
        self.set_header("Content-type", "text/html")
        self.write(rendered_template.encode())

    def get(self, suite_id):
        """Handler for HTTP GET method.

        Renders the suite details HTML page.
        """
        try:
            self._render_suite_details(suite_id)
            self.set_status(200)
        except Exception as e:
            log.error(traceback.format_exc())
            self.set_status(status_code=500, reason=str(e))
        finally:
            self.flush()
            self.finish()


class SuiteStatusHandler(BaseMasterHandler):
    """Handler for requests related to suite experiments."""

    def _get_suite_status(self, suite_id):
        """Writes the JSON-encoded suite status to the output stream.

        :param suite_id: (string) suite id.
        """
        message = {
            "suite_id": suite_id,
            "experiments": [
                {"id": experiment.id, "status": experiment.get_status_string()}
                for experiment in Experiment.get_by_suite(suite_id)
            ],
        }
        self.write(chunk=message)

    def get(self, suite_id):
        """Handler for HTTP GET method.

        Returns information on the suite experiments as a JSON-encoded file.
        """
        try:
            self._get_suite_status(suite_id)
        except Exception as e:
            log.error(traceback.format_exc())
            self.set_status(500, reason=str(e))
        finally:
            self.flush()
            self.finish()


class SuiteDumpHandler(BaseMasterHandler):
    """Handler for requests related to the suite results dump."""

    def _get_suite_dump(self, suite_id):
        """Writes CSV-encoded suite results to the output stream.

        :param suite_id: (string) suite id.
        """

        def get_dump_header(params):
            parameter_names = sorted(params.keys())
            fields = [
                "experiment_id",
                "execution_time_microseconds",
                "data",
                "candidate",
                "roc_auc",
                "average_precision",
                *parameter_names,
            ]
            return ",".join(fields) + "\n"

        def get_digest_line(exp):
            metrics = experiment.load_metrics()

            exp_parameters = load_parameter_string(exp.parameters)
            parameter_values = [v for k, v in sorted(exp_parameters.items())]
            roc_auc = metrics["roc_auc"]
            average_precision = metrics["average_precision"]
            execution_time_microseconds = int(
                (exp.completed_ts - exp.scheduled_ts).total_seconds() * 1e6
            )
            digest_fields = [
                exp.id,
                str(execution_time_microseconds),
                exp.dataset,
                Candidate.get_by_id(exp.candidate).name,
                str(roc_auc),
                str(average_precision),
                *parameter_values,
            ]
            return ",".join(digest_fields) + "\n"

        experiments = Experiment.get_by_suite(suite_id)
        parameters = load_parameter_string(experiments[0].parameters)
        self.write(get_dump_header(parameters))
        for experiment in experiments:
            if experiment.status == ExperimentStatus.COMPLETED:
                experiment_digest_line = get_digest_line(experiment)
                self.write(chunk=experiment_digest_line)

    def get(self, suite_id):
        """Handler for HTTP GET method.

        Returns information on the suite results as a CSV-encoded file.
        """
        try:
            self._get_suite_dump(suite_id)
        except Exception as e:
            log.error(traceback.format_exc())  # Write on master log
            self.set_status(500, reason=str(e))
        finally:
            self.flush()
            self.finish()



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
from pyod.models.feature_bagging import FeatureBagging as FB


class FeatureBagging:
    """Implements the Feature Bagging (FB) method proposed in:

    Lazarevic, Aleksandar, and Vipin Kumar.
    "Feature bagging for outlier detection."
    Proceedings of the eleventh ACM SIGKDD international conference
    on Knowledge discovery in data mining. ACM, 2005.

    This class wraps the implementation provided by the pyod library
    (see https://pyod.readthedocs.io/en/latest/pyod.models.html#module-pyod.models.feature_bagging)  # noqa 501

    Parameters:
    :combination: (string) either 'average' or 'max' (defaults to 'average').
    :m: (int) number of subspaces, i.e. weak detectors (defaults to 10).
    :max_features: (float) fraction of features to use for each
    subspace (defaults to 1.0). Must be between [0.5 and 1.0].
    :k: (int) number of neighbors to use in the weak LOF classifiers (defaults to 10).
    :seed: (int) random number generator seed.
    """

    def __init__(self, **kwargs):
        self.combination = str(kwargs.get("combination", "average"))
        self.estimator_params = {"n_neighbors": int(kwargs.get("k", 10))}
        self.m = int(kwargs.get("m", 10))
        self.max_features = float(kwargs.get("max_features", 1.0))
        self.seed = int(kwargs["seed"])
        self._model = None

    def fit(self, train_data):
        self._model = FB(
            n_estimators=self.m,
            max_features=self.max_features,
            random_state=self.seed,
            combination=self.combination,
            estimator_params=self.estimator_params,
        ).fit(train_data)
        return self

    def score(self, element):
        if not self._model:
            raise ValueError("invalid state. The model has not been trained.")
        element = element.reshape(1, -1)
        return self._model.decision_function(element)[0]



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import os
import subprocess

from bad_framework.bad_utils.files import get_include_dir


def start_bad_master(port, debug=False):
    """Starts the BAD master process on localhost at the given port.

    If debug is True, starts the master in development mode.

    :param port: (int) port to start the server on.
    :param debug: (bool) True if master must be started in
    development mode, False otherwise.
    :return: None
    """
    package_bin_dir = "{include_dir}/bin".format(include_dir=get_include_dir())
    params = {
        "flag": "BAD_DEBUG=true" if debug else "",
        "script": os.path.join(package_bin_dir, "start_master.sh"),
        "port": port,
    }
    subprocess.call(
        ["{flag} source {script} {port} > /dev/null".format(**params)],
        shell=True,
    )


def stop_bad_master():
    """Stops the BAD master process on localhost."""
    package_bin_dir = "{include_dir}/bin".format(include_dir=get_include_dir())
    subprocess.call(
        [
            "source {script} > /dev/null".format(
                script=os.path.join(package_bin_dir, "stop_master.sh")
            )
        ],
        shell=True,
    )



from concurrent.futures import ThreadPoolExecutor, as_completed

from sklearn.neighbors import NearestNeighbors
import pandas as pd
import numpy as np


def _score_partition(k, partition_data):
    """Computes all KNN-outlier scores for elements in the input data.

    :param k:
    :param partition_data:
    :return:
    """
    partition_model = NearestNeighbors(n_neighbors=k, algorithm="kd_tree")
    partition_model.fit(partition_data)

    knn_distances, _ = partition_model.kneighbors()
    scores = np.array([nn_distances[-1] for nn_distances in knn_distances]).reshape(
        (partition_data.shape[0], 1)
    )
    return np.concatenate(
        (partition_data, scores),
        axis=1,
    )


class PartKNN:
    """
    TODO:
     - fix life!
    """

    def __init__(self, **kwargs):
        self.k = int(kwargs.get("k", 10))
        self.partitions_num = int(kwargs.get("partitions_num", 10))
        self.seed = int(kwargs["seed"])
        self._model = None

    def fit(self, train_data):
        """Fits and compute score partition wise.

        Save scores for score method.

        TODO:
         - add partition number to train_matrix
         - create RDD with partition number as partitioning
         - compute KNN partition-wise
         - collect scores and create the model.

        :param train_data:
        :return:
        """
        df = pd.DataFrame(train_data)
        data_size = df.shape[0]

        # Assign a partition to each element
        df["partition"] = np.random.randint(
            low=0, high=self.partitions_num, size=data_size
        )

        # Compute scores independently for each partition
        self._model = {}
        with ThreadPoolExecutor() as executor:
            futures = []
            for partition_index in range(self.partitions_num):
                partition_data = (
                    df[df["partition"] == partition_index]
                    .drop(columns=["partition"])
                    .to_numpy()
                )
                futures.append(
                    executor.submit(
                        _score_partition,
                        k=self.k,
                        partition_data=partition_data,
                    )
                )
            for future in as_completed(futures):
                try:
                    data_with_scores = future.result()
                    for row in data_with_scores:
                        element = tuple(row[:-1])
                        score = row[-1]
                        self._model[element] = score

                except RuntimeError:
                    print("Error computing partition scores")
                    raise RuntimeError
        return self

    def score(self, element):
        """Returns score corresponding to element.

        :param element:
        :return:
        """
        if not self._model:
            raise ValueError("invalid state. The model has not been trained.")
        element_key = tuple(element)
        return self._model.get(element_key, 0.0)


def main():

    rng_seed = 1234
    sample_size = 100
    sample_dims = 2
    partition_size = 5

    np.random.seed(rng_seed)

    sample_params = {
        "k": 2,
        "partition_size": partition_size,
        "seed": rng_seed,
    }

    sample_matrix = np.random.random_sample(size=(sample_size, sample_dims))

    print("Fitting the model...")
    model = PartKNN(**sample_params).fit(sample_matrix)
    print("Done.")

    print("Generating scores...")
    np.apply_along_axis(model.score, axis=1, arr=sample_matrix)
    print("Scores generated correctly.")


if __name__ == "__main__":
    main()



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import numpy as np


class Dummy:
    """Random anomaly detector. Assigns score 1.0 (a.k.a anomalous) to data elements at
    random with probability p.

    Parameters:
    - p: (float) probability of classifying an element as
    anomalous (must be between 0.0 and 1.0).
    - seed: (int) random number generator seed.
    """

    def __init__(self, **kwargs):
        self.p = float(kwargs.get("p", 0.5))
        np.random.seed(int(kwargs["seed"]))

    def fit(self, train_data):
        """No fitting required.

        :param train_data: (numpy.ndarray) training dataset (ignored).
        :return: (bad.Candidate) the trained model.
        """
        return self

    def score(self, element):
        """Classifies element as anomalous with probability p.

        :param element: (numpy.ndarray) data element as a row vector (ignored).
        :return: (float) anomaly score for element.
        """
        random_number = np.random.uniform()
        if random_number >= 1.0 - self.p:
            return 1.0
        return 0.0



import numpy as np

from bad_framework.bad_utils import (
    generate_experiments_settings,
    get_parameter_combinations,
    load_data_matrix,
)
from bad_framework.bad_utils.adt import (
    ExperimentSettings,
    RangeParameter,
    ValueParameter,
)


def test_generate_experiments_settings():
    dummy_datasets = [
        "dummy_data_1",
    ]
    dummy_parameters = {
        "a": ValueParameter(10),
        "b": RangeParameter(2, 6, 2),
    }
    expected_settings = [
        ExperimentSettings("dummy_data_1", "a=10;b=2"),
        ExperimentSettings("dummy_data_1", "a=10;b=4"),
        ExperimentSettings("dummy_data_1", "a=10;b=6"),
    ]
    assert expected_settings == list(
        generate_experiments_settings(dummy_datasets, dummy_parameters)
    )


def test_get_parameter_combinations():
    dummy_parameters = {
        "a": ValueParameter(10),
        "b": RangeParameter(2, 6, 2),
    }
    expected_combinations = [
        (10, 2),
        (10, 4),
        (10, 6),
    ]
    assert expected_combinations == list(get_parameter_combinations(dummy_parameters))


def test_load_data_matrix(tmp_path):
    dummy_data_content = "\n".join(
        [
            "@ RELATION 'Dummy BAD data'",
            "@ ATTRIBUTE 'id' integer",
            "@ ATTRIBUTE 'outlier' real",
            "@ ATTRIBUTE 'att1' real",
            "@ ATTRIBUTE 'att2' real",
            "@ ATTRIBUTE 'att3' real",
            "@ ATTRIBUTE 'att4' real",
            "",
            "@ DATA",
            "0, 0.0, 8.0, 66.0, 9.0, 100.0",
            "1, 1.0, 0.0, 57.0, 33.0, 76.0",
            "2, 0.0, 81.0, 80.0, 28.0, 87.0",
            "3, 1.0, 0.0, 67.0, 28.0, 94.0",
            "4, 0.0, 82.0, 62.0, 85.0, 100.0",
        ]
    )
    data_file = tmp_path / "dummy.arff"
    data_file.write_text(dummy_data_content)
    expected_matrix = np.array(
        [
            [0.0, 0.0, 8.0, 66.0, 9.0, 100.0],
            [1.0, 1.0, 0.0, 57.0, 33.0, 76.0],
            [2.0, 0.0, 81.0, 80.0, 28.0, 87.0],
            [3.0, 1.0, 0.0, 67.0, 28.0, 94.0],
            [4.0, 0.0, 82.0, 62.0, 85.0, 100.0],
        ]
    )
    assert expected_matrix.tolist() == load_data_matrix(data_file).tolist()



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
from uuid import uuid4
import datetime
import glob
import json
import logging
import os

from bad_framework.bad_utils.adt import ExperimentStatus
from bad_framework.bad_utils.files import get_candidate_name, get_include_dir
from bad_framework.bad_utils.network import AsyncHTTPSessionManager

logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)5s - %(message)s", level="INFO"
)
log = logging.getLogger("bad.server.master")


class Model:
    @classmethod
    def get_by_id(cls, query_id):
        return cls._objects[query_id]

    @classmethod
    def get_all(cls):
        return cls._objects.values()

    @classmethod
    def _get_id(cls, tag):
        return tag[:4] + str(uuid4())[:8]


class Candidate(Model):

    _objects = {}

    def __init__(self, suite_id, source_filename, parameters, requirements):
        self.id = self._get_id("candidate")
        self.suite = suite_id
        self.name = get_candidate_name(source_filename)
        self.source = source_filename
        self.parameters = parameters
        self.requirements = requirements

    @classmethod
    def create(cls, suite_id, source_filename, parameters, requirements):
        new_candidate = Candidate(suite_id, source_filename, parameters, requirements)
        cls._objects[new_candidate.id] = new_candidate
        return new_candidate

    @classmethod
    def get_by_suite(cls, suite_id):
        for candidate in cls._objects.values():
            if candidate.suite == suite_id:
                return candidate


class Dataset(Model):

    _objects = {}

    def __init__(self, dataset_name, file_path):
        self.id = self._get_id("dataset")
        self.name = dataset_name
        self.path = file_path

    @classmethod
    def create(cls, dataset_name, file_path):
        new_dataset = Dataset(dataset_name, file_path)
        cls._objects[new_dataset.id] = new_dataset
        return new_dataset

    @classmethod
    def get_by_name(cls, name):
        for dataset in cls._objects.values():
            if dataset.name == name:
                return dataset

    @classmethod
    def setup(cls):
        data_dir = os.path.join(get_include_dir(), "data")
        arff_files = glob.glob("{data_dir}/*.arff".format(data_dir=data_dir))
        for dataset_file in arff_files:
            dataset_name = os.path.splitext(os.path.basename(dataset_file))[0]
            cls.create(dataset_name=dataset_name, file_path=dataset_file)


class Experiment(Model):

    _objects = {}

    _status_strings = {
        ExperimentStatus.CREATED: "created",
        ExperimentStatus.SCHEDULED: "scheduled",
        ExperimentStatus.RUNNING: "running",
        ExperimentStatus.COMPLETED: "completed",
        ExperimentStatus.FAILED: "failed",
    }

    def __init__(self, suite_id, candidate_id, dataset_name, parameters):
        self.id = self._get_id("experiment")
        self.suite = suite_id
        self.candidate = candidate_id
        self.completed_ts = None
        self.dataset = dataset_name
        self.metrics = None
        self.parameters = parameters
        self.roc = None
        self.scheduled_ts = None
        self.scores = None
        self.status = ExperimentStatus.CREATED

    def update_status(self, status):
        if status == ExperimentStatus.RUNNING:
            self.scheduled_ts = datetime.datetime.now()
        if status == ExperimentStatus.COMPLETED:
            self.completed_ts = datetime.datetime.now()
        self.status = status

    def load_metrics(self):
        if self.metrics:
            with open(self.metrics, "r") as metrics_file:
                return json.load(metrics_file)
        else:
            ValueError("metrics file not found.")

    @classmethod
    def create(cls, suite_id, candidate_id, dataset_name, parameters):
        new_experiment = Experiment(suite_id, candidate_id, dataset_name, parameters)
        cls._objects[new_experiment.id] = new_experiment
        return new_experiment

    @classmethod
    def get_by_suite(cls, suite_id):
        suite_experiments = []
        for experiment in cls._objects.values():
            if experiment.suite == suite_id:
                suite_experiments.append(experiment)
        return suite_experiments

    def get_status_string(self):
        return self._status_strings[self.status]


class Suite(Model):

    _objects = {}

    def __init__(self):
        self.id = self._get_id("suite")
        self.created_ts = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    @classmethod
    def create(cls):
        new_suite = Suite()
        cls._objects[new_suite.id] = new_suite
        return new_suite


class Worker(Model):

    _objects = {}

    def __init__(self, hostname, port, master_address):
        self.id = self._get_id("worker")
        self.hostname = hostname
        self.port = int(port)
        self.master_address = master_address
        self.session = AsyncHTTPSessionManager(
            domain="{hostname}:{port}".format(hostname=self.hostname, port=self.port)
        )

    @classmethod
    def create(cls, hostname, port, master_address):
        new_worker = Worker(hostname, port, master_address)
        cls._objects[new_worker.id] = new_worker
        return new_worker

    @classmethod
    def setup(cls, workers, master_address):
        for hostname, port in workers:
            cls.create(hostname, port, master_address)



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.

Package for useful abstract data types (ADTs) for the BAD framework.

CandidateSpec: represents the specification for a Candidate.
CandidateSpec fields are:
- name, candidate name.
- candidate_path, path to Candidate module.
- requirements_path, path to Candidate requirements file.
- parameters_path, path to Candidate parameters file.

>>> cs = CandidateSpec(
...    "local",
...    "DummyCandidate.py",
... )
>>> cs.source
'local'
>>> cs.url
'DummyCandidate.py'

ExperimentSettings: represents the specification of an experiment.
ExperimentSettings fields are:
- data, data set identifier
- parameters, semicolon-separated hyperparameter string.

>>> es = ExperimentSettings("test_data", "a=1;b=2;c=3")
>>> es.dataset_name
'test_data'
>>> es.parameters
'a=1;b=2;c=3'

ValueParameter: represents a parameter specified as a single value.
ValueParameter fields are:
- value, the value for the parameter.

>>> vp = ValueParameter(100)
>>> vp.value
100

RangeParameter: represents a parameter range of values.
RangeParamete fields are:
- start, the first value in the range.
- end, the last value in the range.
- step, the interval between elements in the range.

>>> rp = RangeParameter(2, 10, 2)
>>> rp.start
2
>>> rp.end
10
>>> rp.step
2

RunSpec: represents an assignment of an experiment to a worker.
RunSpec fields are:
- worker, the worker object.
- experiment, the experiment object.

>>> rs = RunSpec("test_worker", "test_experiment")
>>> rs.worker
'test_worker'
>>> rs.experiment
'test_experiment'
"""
from collections import namedtuple

CandidateSpec = namedtuple("CandidateSpec", "source url")
DataSpec = namedtuple("DataSpec", "source url")
ExperimentSettings = namedtuple("ExperimentSettings", "dataset_name parameters")
RangeParameter = namedtuple("RangeParameter", "start end step")
RunSpec = namedtuple("RunSpec", "worker experiment")
ValueParameter = namedtuple("ValueParameter", "value")


class ExperimentStatus:
    CREATED = 0
    SCHEDULED = 1
    RUNNING = 2
    COMPLETED = 3
    FAILED = 4

    @classmethod
    def valid_statuses(cls):
        return [
            "completed",
            "created",
            "failed",
            "running",
            "scheduled",
        ]


def conditional_casting(str_value):
    """Tries to cast a string value to an integer or float.
    Falls back to the most general type if the casting fails.

    :param str_value: (string) value to be casted.
    :return: (int, float or string) the casted input value.

    >>> conditional_casting("1")
    1
    >>> conditional_casting("12a")
    '12a'
    >>> conditional_casting("12.0")
    12.0
    >>> conditional_casting("12a.0")
    '12a.0'
    """
    try:
        try:
            int_value = int(str_value)
            return int_value
        except ValueError:
            float_value = float(str_value)
            return float_value
    except ValueError:
        return str_value



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
from tempfile import NamedTemporaryFile
from threading import Thread
from http.server import HTTPServer, SimpleHTTPRequestHandler
import httpx
import json
import random
import requests

from .files import save_file


class HTTPSessionManager:
    """Utility class for managing a persistent HTTP session."""

    def __init__(self, domain):
        """Initializes the SessionManager to connect to a given hostname.

        :param domain: (string) destination hostname
        """
        if domain.startswith("http://"):
            self._fq_domain = domain
        else:
            self._fq_domain = "http://" + domain

        if httpx.head(self._fq_domain).status_code == 200:
            self._session = requests.Session()
        else:
            raise ValueError("cannot connect to host {}".format(self._fq_domain))

    def download_file(self, url, path=None):
        """GETs the resource specified by URL and saves it in the temporary
        files system folder.

        The file operates in binary mode. The user is responsible for file
        resource deallocation.

        :param url: (string) URL to the resource.
        :param path: (string) optional path to file.
        :return: (string) downloaded file path.
        """
        get_url = "{}/{}".format(self._fq_domain, url)
        r = self._session.get(get_url)
        if r.status_code == 200:
            if path:
                save_file(r.content, path)
                return path
            else:
                with NamedTemporaryFile() as ntf:
                    ntf.write(r.content)
                return ntf.name
        else:
            raise ValueError(r.status_code)

    def get(self, url):
        get_url = "{}/{}".format(self._fq_domain, url)
        return self._session.get(get_url)

    def get_json(self, url):
        get_url = "{}/{}".format(self._fq_domain, url)
        r = self._session.get(get_url)
        if r.status_code == 200:
            return json.loads(r.content)
        raise ValueError(r.status_code)

    def post_json(self, url, data):
        """POSTs JSON data to the specified url

        :param url: (string) resource URL to POST.
        :param data: (dict) JSON data in a python dict.
        :return: (requests.Response) server response.
        """
        post_url = "{}/{}".format(self._fq_domain, url)
        return self._session.post(post_url, json=data)

    def post_files(self, url, files):
        post_url = "{}/{}".format(self._fq_domain, url)
        return self._session.post(post_url, files=files)


class AsyncHTTPSessionManager:
    """Utility class for managing a persistent HTTP session."""

    def __init__(self, domain, timeout=None):
        """Initializes the SessionManager to connect to a given hostname.
        A connection timeout can be optionally specified.

        :param domain: (string) destination hostname.
        :param timeout: (int) connection timeout in seconds (defaults=300, or 5 minutes).
        """
        self._timeout = timeout

        if domain.startswith("http://"):
            self._fq_domain = domain
        else:
            self._fq_domain = "http://" + domain

        if not httpx.head(self._fq_domain).status_code == 200:
            raise ValueError("cannot connect to host {}".format(self._fq_domain))

    async def download_file(self, url, path=None):
        """GETs the resource specified by URL and saves it in the temporary
        files system folder.

        The file operates in binary mode. The user is responsible for file
        resource deallocation.

        :param url: (string) URL to the resource.
        :param path: (string) optional path to file.
        :return: (string) path to the downloaded file.
        """
        get_url = "{}/{}".format(self._fq_domain, url)

        async with httpx.AsyncClient() as client:
            r = await client.get(get_url, timeout=self._timeout)

        if r.status_code == 200:
            if path:
                save_file(r.content, path)
                return path
            else:
                with NamedTemporaryFile() as ntf:
                    ntf.write(r.content)
                return ntf.name
        else:
            raise ValueError(r.status_code)

    async def get(self, url):
        get_url = "{}/{}".format(self._fq_domain, url)
        async with httpx.AsyncClient() as client:
            r = await client.get(get_url, timeout=self._timeout)
        return r

    async def get_json(self, url):
        get_url = "{}/{}".format(self._fq_domain, url)
        async with httpx.AsyncClient() as client:
            r = await client.get(get_url, timeout=self._timeout)
        if r.status_code == 200:
            return json.loads(r.content)
        raise ValueError(r.status_code)

    async def post_json(self, url, data):
        """POSTs JSON data to the specified url

        :param url: (string) resource URL to POST.
        :param data: (dict) JSON data in a python dict.
        :return: (requests.Response) server response.
        """
        post_url = "{}/{}".format(self._fq_domain, url)
        async with httpx.AsyncClient() as client:
            r = await client.post(post_url, json=data, timeout=self._timeout)
        return r

    async def post_files(self, url, files):
        post_url = "{}/{}".format(self._fq_domain, url)
        async with httpx.AsyncClient() as client:
            r = await client.post(post_url, files=files, timeout=self._timeout)
        return r


class MockHTTPHandler(SimpleHTTPRequestHandler):
    """Implements a simple HTTP server configurable to serve arbitrary content.

    This class should be considered as a singleton.
    All instances share the same contents.

    Contents can be added via the serve(obj, at) method and removed
    via the clear() method.

    The serve() method supports all objects that can be dumped using
    the json.dumps(obj) function.
    """

    contents = {}

    def do_GET(self):
        if self.path in self.contents:
            self._serve_contents()
        else:
            self._serve_not_found()

    def _serve_contents(self):
        self.send_response(200)
        self.send_header("Content-Type", "application/json; charset=utf-8")
        self.end_headers()
        self.wfile.write(self.contents[self.path])

    def _serve_not_found(self):
        self.send_response(404)
        self.end_headers()

    @classmethod
    def clear(cls):
        """Clears the content cache served by the MockHTTPHandler class.

        :return:
        """
        cls.contents = {}

    @classmethod
    def serve(cls, obj, at):
        """Configures this server to serve a specific object at a given URL.
        The object is served in the HTTP response body.

        :param obj: (object) object to serve
        :param at: (string) path URL
        :return: (TestServer) this test server, to enable chain calls.
        """
        json_dump = json.dumps(obj)
        cls.contents[at] = bytes(json_dump, encoding="utf-8")
        return cls


class MockServer:
    """Mock HTTP server. In combination with MockHTTPHandler can be set up to
    serve arbitrary objects for testing purposes.

    Supports objects that can be converted to JSON by the json.dumps() function.

    TODO:
     - implement method POST.
    """

    def __init__(self, handler_class, port=None):
        """Initializes the test server with the handler class.
        A port can be optionally defined for the server to listen on.

        If not specified the port is randomly selected in the range [10000, 20000].

        :param handler_class: (SimpleHTTPRequestHandler) handler for the HTTP requests.
        :param port: (int) port for the server to listen on.
        """
        address = ("localhost", port if port else random.randint(10000, 20000))
        self._server = HTTPServer(address, handler_class)
        self._running = False

    def start(self):
        """Starts the test server.

        :return: None
        """
        if not self._running:
            server_thread = Thread(
                target=self._server.serve_forever,
                name="bad-utils_http-test-server",
                daemon=True,
            )
            server_thread.start()
            self._running = True

    def stop(self):
        """Stops the test server.

        :return: None
        """
        if self._running:
            self._server.shutdown()
            self._running = False



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import logging
import os
import subprocess

from bad_framework.bad_utils.files import get_include_dir
import bad_framework

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)5s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
log = logging.getLogger(__name__)


def start_bad_worker(hostname, port, debug=False):

    package_bin_dir = "{include_dir}/bin".format(include_dir=get_include_dir())

    if debug:
        development_flags = "BAD_DEBUG=true BAD_PACKAGE={}".format(
            os.path.dirname(os.path.dirname(bad_framework.__file__))
        )
    else:
        development_flags = ""

    params = {
        "flag": development_flags,
        "script": os.path.join(package_bin_dir, "start_worker.sh"),
        "hostname": hostname,
        "port": port,
    }
    subprocess.call(
        ["{flag} source {script} {hostname} {port} > /dev/null".format(**params)],
        shell=True,
    )


def stop_bad_worker(hostname):
    package_bin_dir = "{include_dir}/bin".format(include_dir=get_include_dir())
    params = {
        "script": os.path.join(package_bin_dir, "stop_worker.sh"),
        "hostname": hostname,
    }
    subprocess.call(
        ["source {script} {hostname} > /dev/null".format(**params)],
        shell=True,
    )



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.

Command-line client for the Benchmarking Anomaly Detection (BAD) framework.
"""
import argparse
import logging
import os
import re
import sys

from bad_framework.bad_utils.files import copy_files_to_bad_directory
from bad_framework.bad_utils.magic import BAD_CONF_DIR, LOG_FORMAT

from bad_framework.bad_client.cli import get_commands, handle_command

logging.basicConfig(
    level=logging.INFO,
    format=LOG_FORMAT,
    datefmt="%Y-%m-%d %H:%M:%S",
)
log = logging.getLogger("bad.client")


def parse_config_file(config_filepath):
    if not os.path.exists(config_filepath):
        raise ValueError("Config file not found: '{}'", config_filepath)
    with open(config_filepath, "r") as config_file:
        key_value_pairs = [
            re.split(r"\s+", line.strip())
            for line in config_file
            if line.strip() and not line.startswith("#")
        ]
        loaded_settings = dict(key_value_pairs)
    return loaded_settings


def add_settings(runtime_config, loaded_settings):
    """Appends the loaded settings to runtime config, unless they are already defined.

    :param runtime_config: (dict) current runtime config.
    :param loaded_settings: (dict) additional config settings.
    :return: (dict) updated runtime config.
    """
    for key in loaded_settings.keys():
        if key not in runtime_config.keys():
            runtime_config[key] = loaded_settings[key]
    return runtime_config


def load_default_config(runtime_config):
    """Loads settings from a configuration file into the runtime configuration
    dictionary.

    Settings already defined in the runtime configuration are not overwritten.

    :param runtime_config: (dict) runtime configuration.
    :return: (dict) updated runtime configuration.
    """
    default_config_path = os.path.join(os.getcwd(), BAD_CONF_DIR)
    if os.path.exists(default_config_path):
        loaded_settings = parse_config_file(default_config_path)
        runtime_config = add_settings(runtime_config, loaded_settings)
    return runtime_config


def parse_arguments():
    """Defines and parses command-line arguments.

    The following positional arguments are required:
     - command

    The returned dictionary always contains the following boolean flag arguments:
     - bad.debug
     - bad.log.verbose

    Optional arguments non present on the command-line are not returned.

    See https://docs.python.org/3/library/argparse.html for details.

    :return: (dict) runtime configuration dictionary.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "command",
        type=str,
        choices=get_commands(),
        help="BAD framework command",
    )
    parser.add_argument(
        "-c",
        "--candidate",
        dest="bad.candidate",
        type=str,
        help="path to Candidate implementation",
    )
    parser.add_argument(
        "-d",
        "--data",
        dest="bad.data",
        type=str,
        help="data set identifier",
    )
    parser.add_argument(
        "-o",
        "--dump-file",
        dest="bad.dump.file",
        type=str,
        help="path to output file",
    )
    parser.add_argument(
        "-p",
        "--parameters",
        dest="bad.candidate.parameters",
        type=str,
        help="path to Candidate parameters file",
    )
    parser.add_argument(
        "-q",
        "--requirements",
        dest="bad.candidate.requirements",
        type=str,
        help="path to Candidate requirements.txt file",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="bad.log.verbose",
        help="enables debug logging",
    )
    parser.add_argument(
        "-D",
        "--dev",
        action="store_true",
        dest="bad.debug",
        help="run the framework in development mode",
    )
    parsed_args = vars(parser.parse_args())
    # Remove None-valued args
    for none_key in [k for k, v in parsed_args.items() if v is None]:
        del parsed_args[none_key]
    return parsed_args


def print_config(config):
    log.debug("Config = {")
    for k, v in config.items():
        log.debug("  %s = %r", k, v)
    log.debug("}")


def initialize_bad_directory():
    cwd = os.getcwd()
    bad_dir = os.path.join(cwd, ".bad")
    if not os.path.exists(bad_dir):
        os.mkdir(bad_dir)
        copy_files_to_bad_directory(bad_dir)


def main():
    config = parse_arguments()
    if config["bad.log.verbose"]:
        log.setLevel("DEBUG")

    initialize_bad_directory()

    config = load_default_config(config)
    print_config(config)

    command = config["command"]
    handle_command(command, config)

    sys.exit(0)


if __name__ == "__main__":
    main()



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
from sklearn.ensemble import IsolationForest


class IForest:
    """Implements the Isolation forest method proposed in:

    Liu, Fei Tony, Kai Ming Ting, and Zhi-Hua Zhou. "Isolation forest."
    Eighth IEEE International Conference on Data Mining. IEEE, 2008.

    Parameters:
    - m: (int) number of trees (defaults to 10).
    - partition_size: (int) number of data elements to use for constructing
    each tree (defaults to 50). Must be smaller than or equal to the data set size.
    - seed: (int) random number generator seed.

    This class wraps the implementation provided by the scikit-learn library
    (see https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.IsolationForest.html).  # noqa 501
    """

    def __init__(self, **kwargs):
        self.m = int(kwargs.get("m", 10))
        self.partition_size = int(kwargs.get("partition_size", 50))
        self.seed = int(kwargs["seed"])
        self._model = None

    def fit(self, train_data):
        self._model = IsolationForest(
            n_estimators=self.m,
            max_samples=self.partition_size,
            behaviour="new",
            contamination="auto",
            random_state=self.seed,
            n_jobs=1,
        ).fit(train_data)
        return self

    def score(self, element):
        if not self._model:
            raise ValueError("invalid state. The model has not been trained.")
        element = element.reshape(1, -1)
        return -self._model.decision_function(element)[0]



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
from pyod.models.ocsvm import OCSVM as SVM


class OCSVM:
    """Implements the one-class Support Vector Machine method proposed in:

    Schölkopf, Bernhard, et al. "Estimating the support of a
    high-dimensional distribution." Neural computation 13.7 (2001):
    1443-1471.

    Parameters:
    - kernel: (string) kernel function to use (defaults to
    'rbf'). It supports 'linear', 'poly', 'rbf', 'sigmoid',
    'precomputed' or a callable.
    - nu: (float) threshold on the number of training
    errors/support vectors to use (defaults to 0.5). Should be
    between (0.0, 1.0].
    - tolerance: (float) error tolerance for the solver (defaults to 1e-3).
    - seed: (int) random number generator seed.

    This class wraps the implementation provided by the pyod
    library (https://pyod.readthedocs.io/en/latest/pyod.models.html#module-pyod.models.ocsvm)  # noqa 501
    """

    def __init__(self, **kwargs):
        self.kernel = str(kwargs.get("kernel", "rbf"))
        self.nu = float(kwargs.get("nu", 0.5))
        self.tolerance = float(kwargs.get("tolerance", 1e-3))
        self.seed = int(kwargs["seed"])
        self._model = None

    def fit(self, train_data):
        self._model = SVM(
            kernel=self.kernel,
            nu=self.nu,
            tol=self.tolerance,
            max_iter=1000,
        ).fit(train_data)
        return self

    def score(self, element):
        if not self._model:
            raise ValueError("invalid state. The model has not been trained.")
        element = element.reshape(1, -1)
        return self._model.decision_function(element)[0]



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import logging

from tornado.options import define, options
import tornado.httpserver
import tornado.ioloop
import tornado.web

from .views import (
    CandidateHandler,
    DatasetHandler,
    ExperimentHandler,
    IndexHandler,
    ResultHandler,
    SuiteHandler,
    SuiteDumpHandler,
    SuiteStatusHandler,
)

# Define tornado options
define(
    "master_home",
    default="/tmp/bad-framework",
    help="Working directory for the master process.",
)
define("master_port", default=3290, help="Port for the master process.")
define("master_debug", default=False, help="Activate development mode.")

# Define our own logging configuration
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)5s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
log = logging.getLogger("bad.server.master")


class BADMasterServer:
    def __init__(self, port, home_dir):
        self._port = int(port)
        self._app = tornado.web.Application(
            [
                (r"^/$", IndexHandler),
                (r"^/index.html$", IndexHandler),
                (r"^/candidate/(?P<candidate_id>[a-z0-9-]+)/$", CandidateHandler),
                (
                    r"^/candidate/(?P<candidate_id>[a-z0-9-]+)/requirements/$",
                    CandidateHandler,
                ),
                (r"^/dataset/(?P<dataset_name>[a-z0-9-]+)/$", DatasetHandler),
                (r"^/experiment/(?P<experiment_id>[a-z0-9-]+)/$", ExperimentHandler),
                (
                    r"^/experiment/(?P<experiment_id>[a-z0-9-]+)/results/$",
                    ResultHandler,
                ),
                (
                    r"^/experiment/(?P<experiment_id>[a-z0-9-]+)/metrics.json$",
                    ResultHandler,
                ),
                (r"^/experiment/(?P<experiment_id>[a-z0-9-]+)/roc.png$", ResultHandler),
                (r"^/suite/$", SuiteHandler),
                (r"^/suite/(?P<suite_id>[a-z0-9-]+)/$", SuiteHandler),
                (r"^/suite/(?P<suite_id>[a-z0-9-]+)/dump/$", SuiteDumpHandler),
                (
                    r"^/suite/(?P<suite_id>[a-z0-9-]+)/status/$",
                    SuiteStatusHandler,
                ),
            ],
            debug=options.master_debug,
            master_home=home_dir,
            master_port=self._port,
        )
        self._server = tornado.httpserver.HTTPServer(
            self._app, max_buffer_size=524288000
        )

    def start(self):
        log.info(">>> Starting BAD master on port %d", self._port)
        self._server.bind(self._port)
        self._server.start(1)
        tornado.ioloop.IOLoop.current().start()


def main():
    tornado.options.parse_command_line()

    log.info("Starting BAD master with the following options:")
    for k, v in options.items():
        log.info("  %s = %r", k, v)

    bad_master = BADMasterServer(options.master_port, options.master_home)
    bad_master.start()


if __name__ == "__main__":
    main()



"""
Test module for the main bad-client module.

Global test fixtures are defined in 'bad-framework/tests/conftest.py'
"""
import pytest
import sys

from bad_framework.bad_client import (
    add_settings,
    load_default_config,
    parse_arguments,
    parse_config_file,
    print_config,
)


def test_add_settings():
    dummy_config = {
        "dummy_one": 1,
        "dummy_two": 2,
    }
    new_settings = {
        "dummy_three": 3,
    }
    expected_config = {
        "dummy_one": 1,
        "dummy_two": 2,
        "dummy_three": 3,
    }
    assert expected_config == add_settings(dummy_config, new_settings)


def test_add_settings_to_empty_config():
    dummy_config = {}
    new_settings = {
        "dummy_one": 1,
    }
    expected_config = {
        "dummy_one": 1,
    }
    assert expected_config == add_settings(dummy_config, new_settings)


def test_add_settings_without_settings():
    dummy_config = {
        "dummy_one": 1,
    }
    new_settings = {}
    expected_config = {
        "dummy_one": 1,
    }
    assert expected_config == add_settings(dummy_config, new_settings)


def test_add_settings_does_not_overwrite():
    dummy_config = {
        "dummy_one": "old_value",
    }
    new_settings = {
        "dummy_one": "new_value",
    }
    expected_config = {
        "dummy_one": "old_value",
    }
    assert expected_config == add_settings(dummy_config, new_settings)


def test_load_default_config(tmp_path):
    import os

    config_file_content = "\n".join(
        ["dummy.path    /config/file/path", "dummy.string  hello", "dummy.value   123"]
    )
    os.chdir(tmp_path)
    bad_dir = tmp_path / ".bad/"
    bad_dir.mkdir()
    config_file = bad_dir / "bad.conf"
    config_file.write_text(config_file_content)

    initial_config = {
        "dummy.path": "original/path",
        "dummy.value": "456",
    }
    expected_config = {
        "dummy.path": "original/path",
        "dummy.string": "hello",
        "dummy.value": "456",
    }
    assert expected_config == load_default_config(initial_config)


def test_parse_arguments_without_arguments(monkeypatch):
    with monkeypatch.context() as mp:
        mp.setattr(sys, "argv", ["bad"])
        with pytest.raises(SystemExit):
            parse_arguments()


def test_parse_arguments_without_command(monkeypatch):
    with monkeypatch.context() as mp:
        mp.setattr(sys, "argv", ["bad", "-c", "./dummy_candidate.py"])
        with pytest.raises(SystemExit):
            parse_arguments()


def test_parse_arguments(monkeypatch):
    with monkeypatch.context() as mp:
        mp.setattr(sys, "argv", ["bad", "run", "-c", "./dummy_candidate.py"])
        expected_config = {
            "command": "run",
            "bad.candidate": "./dummy_candidate.py",
            "bad.debug": False,
            "bad.log.verbose": False,
        }
        assert expected_config == parse_arguments()


def test_parse_arguments_with_debug_flag(monkeypatch):
    with monkeypatch.context() as mp:
        mp.setattr(sys, "argv", ["bad", "run", "-D"])
        expected_config = {
            "command": "run",
            "bad.debug": True,
            "bad.log.verbose": False,
        }
        assert expected_config == parse_arguments()


def test_parse_config_file_not_found():
    with pytest.raises(ValueError):
        parse_config_file("/file/not/found.conf")


def test_parse_empty_config_file(tmp_path):
    config_file = tmp_path / "empty.conf"
    config_file.write_text("")
    assert {} == parse_config_file(config_file.resolve())


def test_parse_config_file(tmp_path):
    config_file_content = "\n".join(
        ["dummy.path    /tmp/hello/world", "dummy.string  hello", "dummy.value   123"]
    )
    config_file = tmp_path / "test.conf"
    config_file.write_text(config_file_content)

    expected_config = {
        "dummy.path": "/tmp/hello/world",
        "dummy.string": "hello",
        "dummy.value": "123",
    }

    assert expected_config == parse_config_file(config_file.resolve())


def test_print_empty_config(bad_caplog):
    empty_config = {}
    expected_messages = ["Config = {", "}"]
    print_config(empty_config)
    assert bad_caplog.messages == expected_messages


def test_print_config(bad_caplog):
    import datetime

    dummy_date = datetime.datetime(1234, 5, 6, 7, 8, 9)

    empty_config = {"hello": "world", "now": dummy_date, "age": 42, "height": 178.9}
    expected_messages = [
        "Config = {",
        "  hello = 'world'",
        "  now = datetime.datetime(1234, 5, 6, 7, 8, 9)",
        "  age = 42",
        "  height = 178.9",
        "}",
    ]
    print_config(empty_config)
    assert bad_caplog.messages == expected_messages



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import datetime
import os


def format_status_message(
    elapsed_seconds, experiments_num, completed_num, failed_num, percentage
):
    elapsed_minutes = elapsed_seconds // 60
    return "Elapsed time: {0}:{1}:{2} - experiments {3}{4}/{5} ({6}%)".format(
        str(elapsed_minutes // 60).zfill(2),
        str(elapsed_minutes % 60).zfill(2),
        str(elapsed_seconds % 60).zfill(2),
        completed_num,
        "" if failed_num == 0 else " ({0} failed)".format(failed_num),
        experiments_num,
        int(percentage * 100),
    )


def get_progress_bar(total_length, percentage, item="#", start="[", end="]"):
    """Builds a progress bar representation, drawn item as base symbol.
    The bar is delimited by start and end.

    The length of the progress bar is proportional to the percentage with respect to
    the total_length.

    :param total_length: (int) total length, representing 100%.
    :param percentage: (float) progress completion percentage, between 0.0 and 1.0.
    :param item: (string) character used to draw the bar.
    :param start: (string) start character for the bar.
    :param end: (string) end character for the bar.
    :return: (string) representing the progress bar.

    Examples:

    >>> get_progress_bar(10, 0.0)
    '[]'
    >>> get_progress_bar(10, 1.0)
    '[########]'
    >>> get_progress_bar(10, 1.0, item='$')
    '[$$$$$$$$]'
    >>> get_progress_bar(10, 0.5, start='E', end='X')
    'E###X'
    """
    bar_length = int(percentage * total_length) - 2
    return start + (item * bar_length) + end


def print_status_bar(start_ts, experiments_num, completed_num, failed_num):
    _, terminal_width = os.popen("stty size", "r").read().split()
    percentage = float(completed_num + failed_num) / experiments_num
    bar = get_progress_bar(int(terminal_width), percentage)
    elapsed_seconds = int((datetime.datetime.now() - start_ts).total_seconds())
    status_message = format_status_message(
        elapsed_seconds,
        experiments_num,
        completed_num,
        failed_num,
        percentage,
    )
    overhead = 4
    if len(status_message) + overhead >= len(bar):
        status_bar = "[ " + status_message + " ]"
    else:
        padded_message = "[ {} ".format(status_message)
        status_bar = padded_message + bar[len(padded_message) :]
    print(status_bar, end="\r", flush=True)



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import datetime
import json
import logging
import os

from bad_framework.bad_master import start_bad_master, stop_bad_master
from bad_framework.bad_worker import start_bad_worker, stop_bad_worker
from bad_framework.bad_utils.adt import CandidateSpec, DataSpec
from bad_framework.bad_utils.files import (
    parse_parameters,
    parse_requirements,
)
from bad_framework.bad_utils.network import HTTPSessionManager

from .monitor import monitor_suite

log = logging.getLogger("bad.client")


def _add_default_parameters(parameters_list, config):
    parameters_list.append(("seed", config["bad.experiment.seed"]))
    parameters_list.append(("trainset_size", config["bad.experiment.trainset_size"]))
    return parameters_list


def _create_suite(config):
    """Hello world!

    TODO:
     - documentation
    :param config:
    :return:
    """
    log.info(">>> Generating experiment suite - BAD master at %s", config["bad.master"])

    suite_settings = _generate_suite_settings(config)

    candidate_spec = suite_settings["candidate"]
    data_spec = suite_settings["data"]

    files = {}

    if candidate_spec.source == "local":
        log.info(">>> Submitting local candidate %s", candidate_spec.url)
        with open(candidate_spec.url, "rb") as local_candidate_file:
            candidate_content = local_candidate_file.read()
        suite_settings["candidate_requirements"] = parse_requirements(
            config["bad.candidate.requirements"]
        )
        files["candidate_source"] = candidate_content

    if data_spec.source == "local":
        log.info(">>> Submitting local dataset %s", data_spec.url)
        with open(data_spec.url, "rb") as local_dataset_file:
            dataset_content = local_dataset_file.read()
        files["data_source"] = dataset_content

    encoded_settings = bytes(json.dumps(suite_settings), encoding="utf-8")
    files["suite_settings"] = encoded_settings

    master_hostname = "{hostname}:{port}".format(
        hostname=config["bad.master"],
        port=config["bad.master.port"],
    )
    master_session = HTTPSessionManager(domain=master_hostname)

    suite_submit_url = "suite/"
    suite_response = master_session.post_files(suite_submit_url, files=files)

    if suite_response.status_code == 200:
        response_message = json.loads(suite_response.content)
        if response_message["status"] == 200:
            log.info("<<< Experiment suite generated correctly.")
            return response_message["suite_id"]
        else:
            raise ValueError(
                "error generating suite - {}".format(response_message["error"])
            )
    else:
        raise ValueError("internal server error {}".format(suite_response.status_code))


def _generate_suite_settings(config):
    """Generates JSON-encoded suite settings to be sent to the master.

    TODO:
     - add support for local data files.
     - add support for remote candidate files.

    :param config:
    :return:
    """
    workers = _load_workers(config["bad.workers"])
    candidate_spec = _load_candidate_spec(config["bad.candidate"])
    data_spec = _load_data_spec(config["bad.data"])
    parameters_list = parse_parameters(config["bad.candidate.parameters"])
    parameters_list = _add_default_parameters(parameters_list, config)

    return {
        "candidate": candidate_spec,
        "candidate_parameters": parameters_list,
        "data": data_spec,
        "master_address": "{master_host}:{master_port}".format(
            master_host=config["bad.master.ip"], master_port=config["bad.master.port"]
        ),
        "workers": workers,
    }


def _download_dump_file(config, suite_id):
    dump_file_path = config["bad.dump.file"]
    log.info(">>> Saving suite dump to %s", dump_file_path)
    session_manager = HTTPSessionManager(
        domain="{master_hostname}:{master_port}".format(
            master_hostname=config["bad.master"],
            master_port=config["bad.master.port"],
        )
    )
    suite_dump_url = "suite/{sid}/dump/".format(sid=suite_id)
    response = session_manager.get(suite_dump_url)

    with open(dump_file_path, "wb+") as dump_file:
        if response.status_code == 200:
            dump_file.write(response.content)
            log.info("<<< Dump file saved.")
        else:
            log.error("%s - %s", response.reason, suite_dump_url)


def _load_candidate_spec(candidate_url):
    if os.path.isfile(candidate_url):
        return CandidateSpec("local", candidate_url)
    else:
        return CandidateSpec("remote", candidate_url)


def _load_data_spec(data_url):
    if os.path.isfile(data_url):
        return DataSpec("local", data_url)
    else:
        return DataSpec("remote", data_url)


def _load_workers(workers_filepath):

    with open(workers_filepath, "r") as workers_file:
        workers = []
        for line in workers_file:
            if line.strip() and not line.startswith("#"):
                worker_host, worker_port = line.strip().split(":")
                workers.append((worker_host, worker_port))
    return workers


def _run_bad_suite(config):

    suite_id = _create_suite(config)

    master_address = "{master_ip}:{master_port}".format(
        master_ip=config["bad.master"], master_port=config["bad.master.port"]
    )
    master_session = HTTPSessionManager(master_address)

    start_time = datetime.datetime.now()
    monitor_suite(master_session=master_session, suite_id=suite_id)
    suite_execution_time = (datetime.datetime.now() - start_time).total_seconds()

    log.info("BAD execution completed in %f seconds.", suite_execution_time)
    _download_dump_file(config, suite_id)


def _start_server(config):
    """Starts the BAD server processes.

    It starts a master process on localhost at the given port.
    It starts a worker process on each worker specified in the workers filepath.

    If debug is true, starts all processes in development mode.

    :param config: (dict) runtime configuration settings.
    :return: None
    """
    master_port = config["bad.master.port"]
    debug = config["bad.debug"]

    log.info(">>> Starting BAD master at localhost on port 3290")
    start_bad_master(master_port, debug)
    log.info("<<< Done.")

    workers_path = config["bad.workers"]
    workers = _load_workers(workers_path)
    for hostname, port in workers:
        log.info(">>> Starting BAD worker at %s on port %s", hostname, port)
        start_bad_worker(hostname, port, debug)
        log.info("<<< Done.")


def _stop_server(config):

    workers = _load_workers(config["bad.workers"])
    for hostname, port in workers:
        log.info(">>> Stopping BAD worker at %s on port %s", hostname, port)
        stop_bad_worker(hostname)
        log.info("<<< Done.")

    log.info(">>> Stopping BAD master at localhost")
    stop_bad_master()
    log.info("<<< Done.")


def _validate_config(command, config):
    """Validates the config settings for a given command. Returns True if the config
    is valid, raises an exception otherwise.

    :param command: (string)
    :param config: (dict)
    """
    required_keys = {
        "run": [
            "bad.candidate",
            "bad.candidate.parameters",
            "bad.data",
            "bad.experiment.seed",
            "bad.experiment.trainset_size",
            "bad.master",
            "bad.master.ip",
            "bad.master.port",
            "bad.workers",
        ],
        "server-start": ["bad.debug", "bad.master.port", "bad.workers"],
        "server-stop": ["bad.workers"],
    }
    for key in required_keys[command]:
        if key not in config:
            raise ValueError("parameter '{key}' is required.".format(key=key))


bad_commands = {
    "run": _run_bad_suite,
    "server-start": _start_server,
    "server-stop": _stop_server,
}


def get_commands():
    """Returns the commands available to the BAD CLI.

    :return: (dict_keys) available commands

    Examples:
    >>> cmds = list(get_commands()); cmds.sort(); cmds
    ['run', 'server-start', 'server-stop']
    """
    return bad_commands.keys()


def handle_command(command, config):
    _validate_config(command, config)
    bad_commands[command](config)



from bad_framework.bad_client.ui import format_status_message


def test_format_status_message():
    expected_message = "Elapsed time: 00:01:15 - experiments 88 (10 failed)/100 (98%)"
    actual_message = format_status_message(
        elapsed_seconds=75,
        experiments_num=100,
        completed_num=88,
        failed_num=10,
        percentage=0.98,
    )
    assert expected_message == actual_message



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import datetime
import logging
import time

from bad_framework.bad_client.ui import print_status_bar
from bad_framework.bad_utils.adt import ExperimentStatus

log = logging.getLogger("bad.client")


def initialize_status_cache(suite_experiments):
    if not suite_experiments:
        raise ValueError("suite not found.")
    status_cache = {}
    for experiment in suite_experiments:
        status_cache[experiment["id"]] = experiment["status"]
    return status_cache


def get_suite_experiments(master_session, suite_id):
    suite_status_url = "suite/{suite_id}/status/".format(suite_id=suite_id)
    suite_status = master_session.get_json(suite_status_url)
    if not suite_status:
        raise ValueError("suite not found.")
    return suite_status["experiments"]


def get_count_by_status(status_cache, query_status):
    if not status_cache:
        raise ValueError("invalid status cache: {}".format(status_cache))
    if query_status not in ExperimentStatus.valid_statuses():
        raise ValueError("invalid status: {}", query_status)
    experiments_with_status = [
        status for status in status_cache.values() if status == query_status
    ]
    return len(experiments_with_status)


def update_status_cache(suite_experiments, status_cache):
    for experiment in suite_experiments:
        if status_cache[experiment["id"]] != experiment["status"]:
            status_cache[experiment["id"]] = experiment["status"]


def monitor_suite(master_session, suite_id, heartbeat_interval=1):
    """Monitors the suite execution from the client. Polls status updates
    at regular intervals and prints a status bar.

    :param master_session:
    :param suite_id: (string) suite identifier.
    :param heartbeat_interval:
    :return: None
    """
    log.info(">>> Starting run monitor")
    suite_experiments = get_suite_experiments(master_session, suite_id)
    status_cache = initialize_status_cache(suite_experiments)

    total = len(status_cache)
    start_ts = datetime.datetime.now()

    while True:
        suite_experiments = get_suite_experiments(master_session, suite_id)
        update_status_cache(suite_experiments, status_cache)
        completed = get_count_by_status(status_cache, "completed")
        failed = get_count_by_status(status_cache, "failed")
        print_status_bar(
            start_ts=start_ts,
            completed_num=completed,
            failed_num=failed,
            experiments_num=total,
        )
        # If the suite is completed break
        if all(
            status == "completed" or status == "failed"
            for status in status_cache.values()
        ):
            print()  # Print a newline to skip over the status bar
            break
        # Otherwise, wait and retry
        time.sleep(heartbeat_interval)
    log.info("<<< Run completed.")



from bad_framework.bad_utils.files import parse_requirements


def test_parse_requirements(tmp_path):
    dummy_requirements = "\n".join(
        [
            "# This is a comment",
            "versioned-requirement-with-spaces >= 1.2.3",
            "versioned-requirement==4.5.6",
            "naked-requirement",
            "requirement-with-comment  # Dummy comment",
            "",
        ]
    )
    requirements_file = tmp_path / "requirements.txt"
    requirements_file.write_text(dummy_requirements)
    expected_requirements = [
        "naked-requirement",
        "requirement-with-comment",
        "versioned-requirement-with-spaces >= 1.2.3",
        "versioned-requirement==4.5.6",
    ]
    assert expected_requirements == parse_requirements(requirements_file)



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
from setuptools import setup, find_packages
import bad_framework

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requirements = [
    "httpx",
    "Jinja2",
    "matplotlib",
    "numpy",
    "pyod",
    "requests",
    "scikit-learn",
    "tornado",
]

setup(
    name="bad-framework",
    version=bad_framework.__version__,
    author="Sivam Pasupathipillai",
    author_email="s.pasupathipillai@unitn.it",
    description="Benchmarking Anomaly Detection (BAD) framework.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        # 'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
    ],
    install_requires=install_requirements,
    python_requires=">=3.6",
    entry_points={"console_scripts": ["bad = bad_framework.bad_client:main"]},
    include_package_data=True,
)



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import os

CANDIDATE_FILES = {
    "dummy": "Dummy.py",
    "fb": "FeatureBagging.py",
    "iforest": "IForest.py",
    "knn": "KNN.py",
    "loci": "LOCI.py",
    "lof": "LOF.py",
    "ocsvm": "OCSVM.py",
    "partknn": "PartKNN.py",
}

CANDIDATE_REQUIREMENTS = {
    "dummy": ["numpy"],
    "fb": ["pyod"],
    "iforest": ["scikit-learn"],
    "knn": ["scikit-learn"],
    "loci": ["pyod"],
    "lof": ["scikit-learn"],
    "ocsvm": ["pyod"],
    "partknn": ["numpy", "pandas", "scikit-learn"],
}


def get_candidate_requirements(candidate_name):
    return CANDIDATE_REQUIREMENTS[candidate_name]


def get_default_candidate(candidate_name):
    """Returns the filename of one of the default candidate modules.

    :param candidate_name:
    :return:
    """
    candidate_dir = os.path.dirname(__file__)
    return os.path.join(candidate_dir, CANDIDATE_FILES[candidate_name])



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
from sklearn.neighbors import NearestNeighbors


class KNN:
    """Implements the KNN outlier detection method proposed in

    Ramaswamy, Sridhar, Rajeev Rastogi, and Kyuseok Shim.
    "Efficient algorithms for mining outliers from large data sets."
    ACM Sigmod Record. Vol. 29. No. 2. ACM, 2000.

    Parameters:
    - k: (int) number of neighbors to consider (defaults to 10). Must
    be smaller than the data set size.
    - seed: (int) random number generator seed.
    """

    def __init__(self, **kwargs):
        self.k = int(kwargs.get("k", 10))
        self._model = None

    def fit(self, train_data):
        self._model = NearestNeighbors(n_neighbors=self.k, algorithm="kd_tree")
        self._model.fit(train_data)
        return self

    def score(self, element):
        if not self._model:
            raise ValueError("invalid state. The model has not been trained.")

        element = element.reshape(1, -1)
        neighbor_distances, _ = self._model.kneighbors(
            element
        )  # neighbor_distances has shape (1, k)
        knn_distance = neighbor_distances[0][-1]
        return knn_distance



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
from tempfile import TemporaryDirectory
import logging
import os
import re
import shutil

from .adt import conditional_casting
import bad_framework

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)5s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
log = logging.getLogger("bad.client")


def get_candidate_name(candidate_path):
    """
    TODO:
     - add error checking (e.g. candidate_path is empty)

    :param candidate_path:
    :return:
    """
    # Matches a python class definition and captures the class name.
    class_name_re = re.compile(r"^class[\s]+(\w+)[(:]")
    candidate_name = None
    with open(candidate_path, "r") as candidate_file:
        for line in candidate_file.readlines():
            match = class_name_re.match(line)
            if match:
                # The first capture group is the class name
                candidate_name = str(match[1])
                break
    return candidate_name


def get_candidate_filename(home_dir, suite_id):
    """Returns a suite-unique filename where to store a local candidate module.

    TODO: refactor, maybe get_local_candidate_filename?

    :param home_dir:
    :param suite_id:
    :return:
    """
    base_path = os.path.join(home_dir, suite_id)
    return os.path.join(base_path, "candidate.py")


def parse_param_line(line):
    """Parse a line of text specifying a parameter and returns the
    parsed fields.

    :param line: (string) line of text to parse
    :return: (list[string]) parameter specification fields.
    """
    return re.split(r"\s+", line.strip())


def parse_value_parameter(fields):
    param_name = fields[0]
    value = conditional_casting(fields[1])
    return param_name, value


def parse_range_parameter(fields):
    param_name = fields[0]
    start = conditional_casting(fields[1])
    end = conditional_casting(fields[2])
    step = conditional_casting(fields[3])
    param_type = start.__class__.__name__
    if type(start) == type(end) == type(step) and (
        param_type == "float" or param_type == "int"
    ):
        return param_name, start, end, step
    else:
        raise ValueError(
            "invalid parameter range for {name}: <{start}, {end}, {step}> ".format(
                name=param_name,
                start=start,
                end=end,
                step=step,
            )
        )


def parse_requirements(requirements_filename):
    requirements = []
    with open(requirements_filename, "r") as requirements_file:
        for line in requirements_file:
            # Remove whitespaces, line and inline comments
            requirement = line.split("#")[0].strip()
            if requirement:
                requirements.append(requirement)
    requirements.sort()
    return requirements


def parse_parameters(parameters_filename):
    parameters = []
    with open(parameters_filename, "r") as parameters_file:
        for line in parameters_file:
            line = line.split("#")[0].strip()
            if line:
                fields = re.split(r"\s+", line)
                num_fields = len(fields)
                if num_fields == 2:
                    parameters.append(parse_value_parameter(fields))
                elif num_fields == 4:
                    parameters.append(parse_range_parameter(fields))
                else:
                    raise ValueError(
                        "invalid parameter specification at: {param_line}".format(
                            param_line=line
                        )
                    )
    return parameters


def get_include_dir():
    return os.path.join(os.path.dirname(bad_framework.__file__), "include")


def save_file(content, path):
    """Saves a byte-array to a binary file path.

    :param content: (bytes) content to write.
    :param path: (string) filepath to write to.
    :return: None
    """
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    with open(path, "wb+") as file:
        file.write(content)


def file_exists_with_content(path, expected_content):
    """Checks that the file identified by path exists and that it contains
    the given content.

    :param path: (string) path to file to check.
    :param expected_content: (bytes) bytes representing the expected content of the file.
    :return: (bool) True if the file exists and has
    the expected content, raises a ValueError otherwise.
    """
    if not os.path.exists(os.path.dirname(path)):
        raise ValueError("file {} does not exist.".format(path))
    with open(path, "rb") as file:
        actual_content = file.read()
        print("Actual content is:", actual_content)
        if not actual_content == expected_content:
            raise ValueError(
                "file {} does not contain the expected content.".format(path)
            )
    return True


def get_init_paths():
    """Returns the list of files that must be present in order to run BAD
    correctly.

    :return: (list[string]) list of paths

    >>> paths = get_init_paths(); paths.sort(); paths
    ['bad.conf', 'candidate_parameters.txt', 'candidate_requirements.txt', 'workers']
    """
    package_init_dir = os.path.join(get_include_dir(), "defaults")
    raw_files = os.listdir(package_init_dir)
    exclude_files = [
        "__pycache__",
    ]
    return [file for file in raw_files if file not in exclude_files]


def copy_files_to_bad_directory(dest_dir):
    """Initializes the directory with all required files to run the BAD framework.
    These include all files in the bad_framework/include/defaults directory.
    """

    # Copy all contents from the package to a temporary dir.
    src_dir_path = os.path.join(get_include_dir(), "defaults")

    with TemporaryDirectory() as temp_dir:
        dest_dir_path = os.path.join(temp_dir, "buffer_dir")
        log.debug("Copying all contents from %s to %s", src_dir_path, dest_dir_path)
        shutil.copytree(src_dir_path, dest_dir_path)

        # Copy all contents of the temporary directory to the current directory
        for relative_path in get_init_paths():
            absolute_src = os.path.join(dest_dir_path, relative_path)
            if os.path.isfile(absolute_src):
                shutil.copy(absolute_src, dest_dir)
            elif os.path.isdir(absolute_src):
                shutil.copytree(absolute_src, os.path.join(dest_dir, relative_path))
            else:
                raise ValueError("neither a file nor a directory.")



"""Copyright (C) 2020 Sivam Pasupathipillai <s.pasupathipillai@unitn.it>.

All rights reserved.
"""
import os

BAD_DIR = ".bad"
BAD_CONF_DIR = os.path.join(BAD_DIR, "bad.conf")
LOG_FORMAT = "%(asctime)s - %(name)s - %(levelname)5s - %(message)s"
